@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2012-2015] Novell, Inc.
- * Copyright (c) [2016-2021] SUSE LLC
+ * Copyright (c) [2016-2022] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -38,17 +38,22 @@
 boost::shared_mutex big_mutex;
 
 
-Client::Client(const string& name, const Clients& clients)
-    : name(name), clients(clients)
+Client::Client(const string& name, uid_t uid, const Clients& clients)
+    : name(name), uid(uid), clients(clients)
 {
 }
 
 
 Client::~Client()
 {
-    thread.interrupt();
-    if (thread.joinable())
-	thread.join();
+    method_call_thread.interrupt();
+    files_transfer_thread.interrupt();
+
+    if (method_call_thread.joinable())
+	method_call_thread.join();
+
+    if (files_transfer_thread.joinable())
+	files_transfer_thread.join();
 
     for (list<Comparison*>::iterator it = comparisons.begin(); it != comparisons.end(); ++it)
     {
@@ -369,6 +374,13 @@ Client::introspect(DBus::Connection& conn, DBus::Message& msg)
 	"      <arg name='files' type='a(su)' direction='out'/>\n"
 	"    </method>\n"
 
+	"    <method name='GetFilesByPipe'>\n"
+	"      <arg name='config-name' type='s' direction='in'/>\n"
+	"      <arg name='number1' type='u' direction='in'/>\n"
+	"      <arg name='number2' type='u' direction='in'/>\n"
+	"      <arg name='fd' type='h' direction='out'/>\n"
+	"    </method>\n"
+
 	"    <method name='Sync'>\n"
 	"      <arg name='config-name' type='s' direction='in'/>\n"
 	"    </method>\n"
@@ -378,8 +390,8 @@ Client::introspect(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << introspect;
+    DBus::Marshaller marshaller(reply);
+    marshaller << introspect;
 
     conn.send(reply);
 }
@@ -394,7 +406,7 @@ struct Permissions : public Exception
 void
 Client::check_permission(DBus::Connection& conn, DBus::Message& msg) const
 {
-    unsigned long uid = conn.get_unix_userid(msg);
+    // Check if the uid of the dbus-user is root.
     if (uid == 0)
 	return;
 
@@ -406,14 +418,12 @@ void
 Client::check_permission(DBus::Connection& conn, DBus::Message& msg,
 			 const MetaSnapper& meta_snapper) const
 {
-    unsigned long uid = conn.get_unix_userid(msg);
-
     // Check if the uid of the dbus-user is root.
     if (uid == 0)
 	return;
 
     // Check if the uid of the dbus-user is included in the allowed uids.
-    if (contains(meta_snapper.uids, uid))
+    if (contains(meta_snapper.get_allowed_uids(), uid))
 	return;
 
     string username;
@@ -422,7 +432,7 @@ Client::check_permission(DBus::Connection& conn, DBus::Message& msg,
     if (get_uid_username_gid(uid, username, gid))
     {
 	// Check if the primary gid of the dbus-user is included in the allowed gids.
-	if (contains(meta_snapper.gids, gid))
+	if (contains(meta_snapper.get_allowed_gids(), gid))
 	    return;
 
 	vector<gid_t> gids = getgrouplist(username.c_str(), gid);
@@ -430,7 +440,7 @@ Client::check_permission(DBus::Connection& conn, DBus::Message& msg,
 	// Check if any (primary or secondary) gid of the dbus-user is included in the allowed
 	// gids.
 	for (vector<gid_t>::const_iterator it = gids.begin(); it != gids.end(); ++it)
-	    if (contains(meta_snapper.gids, *it))
+	    if (contains(meta_snapper.get_allowed_gids(), *it))
 		return;
     }
 
@@ -497,8 +507,8 @@ Client::signal_config_created(DBus::Connection& conn, const string& config_name)
 {
     DBus::MessageSignal msg(PATH, INTERFACE, "ConfigCreated");
 
-    DBus::Hoho hoho(msg);
-    hoho << config_name;
+    DBus::Marshaller marshaller(msg);
+    marshaller << config_name;
 
     conn.send(msg);
 }
@@ -509,8 +519,8 @@ Client::signal_config_modified(DBus::Connection& conn, const string& config_name
 {
     DBus::MessageSignal msg(PATH, INTERFACE, "ConfigModified");
 
-    DBus::Hoho hoho(msg);
-    hoho << config_name;
+    DBus::Marshaller marshaller(msg);
+    marshaller << config_name;
 
     conn.send(msg);
 }
@@ -521,8 +531,8 @@ Client::signal_config_deleted(DBus::Connection& conn, const string& config_name)
 {
     DBus::MessageSignal msg(PATH, INTERFACE, "ConfigDeleted");
 
-    DBus::Hoho hoho(msg);
-    hoho << config_name;
+    DBus::Marshaller marshaller(msg);
+    marshaller << config_name;
 
     conn.send(msg);
 }
@@ -534,8 +544,8 @@ Client::signal_snapshot_created(DBus::Connection& conn, const string& config_nam
 {
     DBus::MessageSignal msg(PATH, INTERFACE, "SnapshotCreated");
 
-    DBus::Hoho hoho(msg);
-    hoho << config_name << num;
+    DBus::Marshaller marshaller(msg);
+    marshaller << config_name << num;
 
     conn.send(msg);
 }
@@ -547,8 +557,8 @@ Client::signal_snapshot_modified(DBus::Connection& conn, const string& config_na
 {
     DBus::MessageSignal msg(PATH, INTERFACE, "SnapshotModified");
 
-    DBus::Hoho hoho(msg);
-    hoho << config_name << num;
+    DBus::Marshaller marshaller(msg);
+    marshaller << config_name << num;
 
     conn.send(msg);
 }
@@ -560,8 +570,8 @@ Client::signal_snapshots_deleted(DBus::Connection& conn, const string& config_na
 {
     DBus::MessageSignal msg(PATH, INTERFACE, "SnapshotsDeleted");
 
-    DBus::Hoho hoho(msg);
-    hoho << config_name << nums;
+    DBus::Marshaller marshaller(msg);
+    marshaller << config_name << nums;
 
     conn.send(msg);
 }
@@ -576,11 +586,11 @@ Client::list_configs(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho.open_array(DBus::TypeInfo<ConfigInfo>::signature);
+    DBus::Marshaller marshaller(reply);
+    marshaller.open_array(DBus::TypeInfo<ConfigInfo>::signature);
     for (MetaSnappers::const_iterator it = meta_snappers.begin(); it != meta_snappers.end(); ++it)
-	hoho << it->getConfigInfo();
-    hoho.close_array();
+    marshaller << it->getConfigInfo();
+    marshaller.close_array();
 
     conn.send(reply);
 }
@@ -591,8 +601,8 @@ Client::get_config(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("GetConfig config_name:" << config_name);
 
@@ -604,8 +614,8 @@ Client::get_config(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << it->getConfigInfo();
+    DBus::Marshaller marshaller(reply);
+    marshaller << it->getConfigInfo();
 
     conn.send(reply);
 }
@@ -616,9 +626,9 @@ Client::set_config(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
+    DBus::Unmarshaller unmarshaller(msg);
     map<string, string> raw;
-    hihi >> config_name >> raw;
+    unmarshaller >> config_name >> raw;
 
     y2deb("SetConfig config_name:" << config_name << " raw:" << raw);
 
@@ -646,8 +656,8 @@ Client::create_config(DBus::Connection& conn, DBus::Message& msg)
     string fstype;
     string template_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> subvolume >> fstype >> template_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> subvolume >> fstype >> template_name;
 
     y2deb("CreateConfig config_name:" << config_name << " subvolume:" << subvolume <<
 	  " fstype:" << fstype << " template_name:" << template_name);
@@ -671,8 +681,8 @@ Client::delete_config(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("DeleteConfig config_name:" << config_name);
 
@@ -699,8 +709,8 @@ Client::lock_config(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("LockConfig config_name:" << config_name);
 
@@ -723,8 +733,8 @@ Client::unlock_config(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("UnlockConfig config_name:" << config_name);
 
@@ -747,8 +757,8 @@ Client::list_snapshots(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("ListSnapshots config_name:" << config_name);
 
@@ -763,8 +773,8 @@ Client::list_snapshots(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << snapshots;
+    DBus::Marshaller marshaller(reply);
+    marshaller << snapshots;
 
     conn.send(reply);
 }
@@ -776,8 +786,8 @@ Client::list_snapshots_at_time(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     time_t begin, end;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> begin >> end;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> begin >> end;
 
     y2deb("ListSnapshotsAtTime config_name:" << config_name << " begin:" << begin <<
 	  " end:" << end);
@@ -793,15 +803,15 @@ Client::list_snapshots_at_time(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
+    DBus::Marshaller marshaller(reply);
 
-    hoho.open_array(DBus::TypeInfo<Snapshot>::signature);
+    marshaller.open_array(DBus::TypeInfo<Snapshot>::signature);
     for (Snapshots::const_iterator it = snapshots.begin(); it != snapshots.end(); ++it)
     {
 	if (it->getDate() >= begin && it->getDate() <= end)
-	    hoho << *it;
+	    marshaller << *it;
     }
-    hoho.close_array();
+    marshaller.close_array();
 
     conn.send(reply);
 }
@@ -813,8 +823,8 @@ Client::get_snapshot(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     dbus_uint32_t num;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num;
 
     y2deb("GetSnapshot config_name:" << config_name << " num:" << num);
 
@@ -833,8 +843,8 @@ Client::get_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << *snap;
+    DBus::Marshaller marshaller(reply);
+    marshaller << *snap;
 
     conn.send(reply);
 }
@@ -847,8 +857,8 @@ Client::set_snapshot(DBus::Connection& conn, DBus::Message& msg)
     dbus_uint32_t num;
     SMD smd;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num >> smd.description >> smd.cleanup >> smd.userdata;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num >> smd.description >> smd.cleanup >> smd.userdata;
 
     y2deb("SetSnapshot config_name:" << config_name << " num:" << num);
 
@@ -881,8 +891,8 @@ Client::create_single_snapshot(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     SCD scd;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> scd.description >> scd.cleanup >> scd.userdata;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> scd.description >> scd.cleanup >> scd.userdata;
 
     y2deb("CreateSingleSnapshot config_name:" << config_name << " description:" << scd.description <<
 	  " cleanup:" << scd.cleanup);
@@ -892,7 +902,7 @@ Client::create_single_snapshot(DBus::Connection& conn, DBus::Message& msg)
     MetaSnappers::iterator it = meta_snappers.find(config_name);
 
     check_permission(conn, msg, *it);
-    scd.uid = conn.get_unix_userid(msg);
+    scd.uid = uid;
 
     Snapper* snapper = it->getSnapper();
 
@@ -900,8 +910,8 @@ Client::create_single_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << snap1->getNum();
+    DBus::Marshaller marshaller(reply);
+    marshaller << snap1->getNum();
 
     conn.send(reply);
 
@@ -916,8 +926,8 @@ Client::create_single_snapshot_v2(DBus::Connection& conn, DBus::Message& msg)
     unsigned int parent_num;
     SCD scd;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> parent_num >> scd.read_only >> scd.description >> scd.cleanup >> scd.userdata;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> parent_num >> scd.read_only >> scd.description >> scd.cleanup >> scd.userdata;
 
     y2deb("CreateSingleSnapshotV2 config_name:" << config_name << " parent_num:" << parent_num <<
 	  " read_only:" << scd.read_only << " description:" << scd.description << " cleanup:" << scd.cleanup);
@@ -927,7 +937,7 @@ Client::create_single_snapshot_v2(DBus::Connection& conn, DBus::Message& msg)
     MetaSnappers::iterator it = meta_snappers.find(config_name);
 
     check_permission(conn, msg, *it);
-    scd.uid = conn.get_unix_userid(msg);
+    scd.uid = uid;
 
     Snapper* snapper = it->getSnapper();
 
@@ -939,8 +949,8 @@ Client::create_single_snapshot_v2(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << snap2->getNum();
+    DBus::Marshaller marshaller(reply);
+    marshaller << snap2->getNum();
 
     conn.send(reply);
 
@@ -954,8 +964,8 @@ Client::create_single_snapshot_of_default(DBus::Connection& conn, DBus::Message&
     string config_name;
     SCD scd;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> scd.read_only >> scd.description >> scd.cleanup >> scd.userdata;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> scd.read_only >> scd.description >> scd.cleanup >> scd.userdata;
 
     y2deb("CreateSingleSnapshotOfDefault config_name:" << config_name << " read_only:" <<
 	  scd.read_only << " description:" << scd.description << " cleanup:" << scd.cleanup);
@@ -965,7 +975,7 @@ Client::create_single_snapshot_of_default(DBus::Connection& conn, DBus::Message&
     MetaSnappers::iterator it = meta_snappers.find(config_name);
 
     check_permission(conn, msg, *it);
-    scd.uid = conn.get_unix_userid(msg);
+    scd.uid = uid;
 
     Snapper* snapper = it->getSnapper();
 
@@ -973,8 +983,8 @@ Client::create_single_snapshot_of_default(DBus::Connection& conn, DBus::Message&
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << snap->getNum();
+    DBus::Marshaller marshaller(reply);
+    marshaller << snap->getNum();
 
     conn.send(reply);
 
@@ -988,8 +998,8 @@ Client::create_pre_snapshot(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     SCD scd;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> scd.description >> scd.cleanup >> scd.userdata;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> scd.description >> scd.cleanup >> scd.userdata;
 
     y2deb("CreatePreSnapshot config_name:" << config_name << " description:" << scd.description <<
 	  " cleanup:" << scd.cleanup);
@@ -999,7 +1009,7 @@ Client::create_pre_snapshot(DBus::Connection& conn, DBus::Message& msg)
     MetaSnappers::iterator it = meta_snappers.find(config_name);
 
     check_permission(conn, msg, *it);
-    scd.uid = conn.get_unix_userid(msg);
+    scd.uid = uid;
 
     Snapper* snapper = it->getSnapper();
 
@@ -1007,8 +1017,8 @@ Client::create_pre_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << snap1->getNum();
+    DBus::Marshaller marshaller(reply);
+    marshaller << snap1->getNum();
 
     conn.send(reply);
 
@@ -1023,8 +1033,8 @@ Client::create_post_snapshot(DBus::Connection& conn, DBus::Message& msg)
     unsigned int pre_num;
     SCD scd;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> pre_num >> scd.description >> scd.cleanup >> scd.userdata;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> pre_num >> scd.description >> scd.cleanup >> scd.userdata;
 
     y2deb("CreatePostSnapshot config_name:" << config_name << " pre_num:" << pre_num <<
 	  " description:" << scd.description << " cleanup:" << scd.cleanup);
@@ -1034,7 +1044,7 @@ Client::create_post_snapshot(DBus::Connection& conn, DBus::Message& msg)
     MetaSnappers::iterator it = meta_snappers.find(config_name);
 
     check_permission(conn, msg, *it);
-    scd.uid = conn.get_unix_userid(msg);
+    scd.uid = uid;
 
     Snapper* snapper = it->getSnapper();
     Snapshots& snapshots = snapper->getSnapshots();
@@ -1044,14 +1054,14 @@ Client::create_post_snapshot(DBus::Connection& conn, DBus::Message& msg)
     Snapshots::iterator snap2 = snapper->createPostSnapshot(snap1, scd);
 
     bool background_comparison = true;
-    it->getConfigInfo().getValue("BACKGROUND_COMPARISON", background_comparison);
+    it->getConfigInfo().get_value("BACKGROUND_COMPARISON", background_comparison);
     if (background_comparison)
 	clients.backgrounds().add_task(it, snap1, snap2);
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << snap2->getNum();
+    DBus::Marshaller marshaller(reply);
+    marshaller << snap2->getNum();
 
     conn.send(reply);
 
@@ -1065,8 +1075,8 @@ Client::delete_snapshots(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     list<dbus_uint32_t> nums;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> nums;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> nums;
 
     y2deb("DeleteSnapshots config_name:" << config_name << " nums:" << nums);
 
@@ -1103,8 +1113,8 @@ Client::get_default_snapshot(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("GetDefaultSnapshot config_name:" << config_name);
 
@@ -1121,12 +1131,12 @@ Client::get_default_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
+    DBus::Marshaller marshaller(reply);
 
     if (tmp != snapshots.end())
-	hoho << true << tmp->getNum();
+    marshaller << true << tmp->getNum();
     else
-	hoho << false << (unsigned int)(0);
+    marshaller << false << (unsigned int)(0);
 
     conn.send(reply);
 }
@@ -1137,8 +1147,8 @@ Client::get_active_snapshot(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("GetActiveSnapshot config_name:" << config_name);
 
@@ -1155,12 +1165,12 @@ Client::get_active_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
+    DBus::Marshaller marshaller(reply);
 
     if (tmp != snapshots.end())
-	hoho << true << tmp->getNum();
+    marshaller << true << tmp->getNum();
     else
-	hoho << false << (unsigned int)(0);
+    marshaller << false << (unsigned int)(0);
 
     conn.send(reply);
 }
@@ -1171,8 +1181,8 @@ Client::calculate_used_space(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("CalculateUsedSpace config_name:" << config_name);
 
@@ -1198,8 +1208,8 @@ Client::get_used_space(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     dbus_uint32_t num;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num;
 
     y2deb("GetUsedSpace config_name:" << config_name << " num:" << num);
 
@@ -1221,8 +1231,8 @@ Client::get_used_space(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << used_space;
+    DBus::Marshaller marshaller(reply);
+    marshaller << used_space;
 
     conn.send(reply);
 }
@@ -1235,8 +1245,8 @@ Client::mount_snapshot(DBus::Connection& conn, DBus::Message& msg)
     dbus_uint32_t num;
     bool user_request;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num >> user_request;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num >> user_request;
 
     y2deb("MountSnapshot config_name:" << config_name << " num:" << num <<
 	  " user_request:" << user_request);
@@ -1263,8 +1273,8 @@ Client::mount_snapshot(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << mount_point;
+    DBus::Marshaller marshaller(reply);
+    marshaller << mount_point;
 
     conn.send(reply);
 }
@@ -1277,8 +1287,8 @@ Client::umount_snapshot(DBus::Connection& conn, DBus::Message& msg)
     dbus_uint32_t num;
     bool user_request;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num >> user_request;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num >> user_request;
 
     y2deb("UmountSnapshot config_name:" << config_name << " num:" << num <<
 	  " user_request:" << user_request);
@@ -1313,8 +1323,8 @@ Client::get_mount_point(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     dbus_uint32_t num;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num;
 
     y2deb("GetMountPoint config_name:" << config_name << " num:" << num);
 
@@ -1334,8 +1344,8 @@ Client::get_mount_point(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << mount_point;
+    DBus::Marshaller marshaller(reply);
+    marshaller << mount_point;
 
     conn.send(reply);
 }
@@ -1347,8 +1357,8 @@ Client::create_comparison(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     dbus_uint32_t num1, num2;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num1 >> num2;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num1 >> num2;
 
     y2deb("CreateComparison config_name:" << config_name << " num1:" << num1 << " num2:" << num2);
 
@@ -1377,9 +1387,9 @@ Client::create_comparison(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
+    DBus::Marshaller marshaller(reply);
     dbus_uint32_t num_files = comparison->getFiles().size();
-    hoho << num_files;
+    marshaller << num_files;
 
     conn.send(reply);
 }
@@ -1391,8 +1401,8 @@ Client::delete_comparison(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     dbus_uint32_t num1, num2;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num1 >> num2;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num1 >> num2;
 
     y2deb("DeleteComparison config_name:" << config_name << " num1:" << num1 << " num2:" << num2);
 
@@ -1419,8 +1429,8 @@ Client::get_files(DBus::Connection& conn, DBus::Message& msg)
     string config_name;
     dbus_uint32_t num1, num2;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name >> num1 >> num2;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num1 >> num2;
 
     y2deb("GetFiles config_name:" << config_name << " num1:" << num1 << " num2:" << num2);
 
@@ -1436,20 +1446,56 @@ Client::get_files(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << files;
+    DBus::Marshaller marshaller(reply);
+    marshaller << files;
 
     conn.send(reply);
 }
 
 
+void
+Client::get_files_by_pipe(DBus::Connection& conn, DBus::Message& msg)
+{
+    string config_name;
+    dbus_uint32_t num1, num2;
+
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name >> num1 >> num2;
+
+    y2deb("GetFilesByPipe config_name:" << config_name << " num1:" << num1 << " num2:" << num2);
+
+    boost::unique_lock<boost::shared_mutex> lock(big_mutex);
+
+    MetaSnappers::iterator it = meta_snappers.find(config_name);
+
+    check_permission(conn, msg, *it);
+
+    list<Comparison*>::iterator it2 = find_comparison(it->getSnapper(), num1, num2);
+
+    const Files& files = (*it2)->getFiles();
+
+    DBus::MessageMethodReturn reply(msg);
+
+    DBus::Marshaller marshaller(reply);
+
+    shared_ptr<FilesTransferTask> files_transfer_task = make_shared<FilesTransferTask>(files);
+
+    marshaller << files_transfer_task->get_read_end();
+    conn.send(reply);
+
+    files_transfer_task->get_read_end().close();
+
+    add_files_transfer_task(files_transfer_task);
+}
+
+
 void
 Client::setup_quota(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("SetupQuota config_name:" << config_name);
 
@@ -1474,8 +1520,8 @@ Client::prepare_quota(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("PrepareQuota config_name:" << config_name);
 
@@ -1500,8 +1546,8 @@ Client::query_quota(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("QueryQuota config_name:" << config_name);
 
@@ -1517,8 +1563,8 @@ Client::query_quota(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << quota_data;
+    DBus::Marshaller marshaller(reply);
+    marshaller << quota_data;
 
     conn.send(reply);
 }
@@ -1529,8 +1575,8 @@ Client::query_free_space(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("QueryFreeSpace config_name:" << config_name);
 
@@ -1546,8 +1592,8 @@ Client::query_free_space(DBus::Connection& conn, DBus::Message& msg)
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
-    hoho << free_space_data;
+    DBus::Marshaller marshaller(reply);
+    marshaller << free_space_data;
 
     conn.send(reply);
 }
@@ -1558,8 +1604,8 @@ Client::sync(DBus::Connection& conn, DBus::Message& msg)
 {
     string config_name;
 
-    DBus::Hihi hihi(msg);
-    hihi >> config_name;
+    DBus::Unmarshaller unmarshaller(msg);
+    unmarshaller >> config_name;
 
     y2deb("Sync config_name:" << config_name);
 
@@ -1588,15 +1634,22 @@ Client::debug(DBus::Connection& conn, DBus::Message& msg) const
 
     DBus::MessageMethodReturn reply(msg);
 
-    DBus::Hoho hoho(reply);
+    DBus::Marshaller marshaller(reply);
 
-    hoho.open_array("s");
+    marshaller.open_array("s");
 
-    hoho << "clients:";
+    marshaller << "server:";
+    {
+	std::ostringstream s;
+	s << "    pid:" << getpid();
+    marshaller << s.str();
+    }
+
+    marshaller << "clients:";
     for (Clients::const_iterator it = clients.begin(); it != clients.end(); ++it)
     {
 	std::ostringstream s;
-	s << "    name:'" << it->name << "'";
+	s << "    name:'" << it->name << "', uid:" << it->uid;
 	if (&*it == this)
 	    s << ", myself";
 	if (it->zombie)
@@ -1605,18 +1658,18 @@ Client::debug(DBus::Connection& conn, DBus::Message& msg) const
 	    s << ", locks " << it->locks.size();
 	if (!it->comparisons.empty())
 	    s << ", comparisons " << it->comparisons.size();
-	hoho << s.str();
+    marshaller << s.str();
     }
 
-    hoho << "backgrounds:";
+    marshaller << "backgrounds:";
     for (Backgrounds::const_iterator it = clients.backgrounds().begin(); it != clients.backgrounds().end(); ++it)
     {
 	std::ostringstream s;
 	s << "    name:'" << it->meta_snapper->configName() << "'";
-	hoho << s.str();
+    marshaller << s.str();
     }
 
-    hoho << "meta-snappers:";
+    marshaller << "meta-snappers:";
     for (MetaSnappers::const_iterator it = meta_snappers.begin(); it != meta_snappers.end(); ++it)
     {
 	std::ostringstream s;
@@ -1629,14 +1682,14 @@ Client::debug(DBus::Connection& conn, DBus::Message& msg) const
 	    else
 		s << ", use count " << it->use_count();
 	}
-	hoho << s.str();
+    marshaller << s.str();
     }
 
-    hoho << "compile options:";
-    hoho << "    version " + string(Snapper::compileVersion());
-    hoho << "    flags " + string(Snapper::compileFlags());
+    marshaller << "compile options:";
+    marshaller << "    version " + string(Snapper::compileVersion());
+    marshaller << "    flags " + string(Snapper::compileFlags());
 
-    hoho.close_array();
+    marshaller.close_array();
 
     conn.send(reply);
 }
@@ -1701,6 +1754,8 @@ Client::dispatch(DBus::Connection& conn, DBus::Message& msg)
 	    delete_comparison(conn, msg);
 	else if (msg.is_method_call(INTERFACE, "GetFiles"))
 	    get_files(conn, msg);
+	else if (msg.is_method_call(INTERFACE, "GetFilesByPipe"))
+	    get_files_by_pipe(conn, msg);
 	else if (msg.is_method_call(INTERFACE, "SetupQuota"))
 	    setup_quota(conn, msg);
 	else if (msg.is_method_call(INTERFACE, "PrepareQuota"))
@@ -1879,6 +1934,12 @@ Client::dispatch(DBus::Connection& conn, DBus::Message& msg)
 	DBus::MessageError reply(msg, "error.unsupported", e.what());
 	conn.send(reply);
     }
+    catch (const StreamException& e)
+    {
+	SN_CAUGHT(e);
+	DBus::MessageError reply(msg, "error.stream", DBUS_ERROR_FAILED);
+	conn.send(reply);
+    }
     catch (const Exception& e)
     {
 	SN_CAUGHT(e);
@@ -1895,34 +1956,48 @@ Client::dispatch(DBus::Connection& conn, DBus::Message& msg)
 
 
 void
-Client::add_task(DBus::Connection& conn, DBus::Message& msg)
+Client::add_method_call_task(DBus::Connection& conn, DBus::Message& msg)
 {
-    if (thread.get_id() == boost::thread::id())
-	thread = boost::thread(boost::bind(&Client::worker, this));
+    if (method_call_thread.get_id() == boost::thread::id())
+	method_call_thread = boost::thread(boost::bind(&Client::method_call_worker, this));
 
-    boost::unique_lock<boost::mutex> lock(mutex);
-    tasks.push(Task(conn, msg));
+    boost::unique_lock<boost::mutex> lock(method_call_mutex);
+    method_call_tasks.push(MethodCallTask(conn, msg));
     lock.unlock();
 
-    condition.notify_one();
+    method_call_condition.notify_one();
 }
 
 
 void
-Client::worker()
+Client::add_files_transfer_task(shared_ptr<FilesTransferTask> files_transfer_task)
+{
+    if (files_transfer_thread.get_id() == boost::thread::id())
+	files_transfer_thread = boost::thread(boost::bind(&Client::files_transfer_worker, this));
+
+    boost::unique_lock<boost::mutex> lock(files_transfer_mutex);
+    files_transfer_tasks.push(files_transfer_task);
+    lock.unlock();
+
+    files_transfer_condition.notify_one();
+}
+
+
+void
+Client::method_call_worker()
 {
     try
     {
 	while (true)
 	{
-	    boost::unique_lock<boost::mutex> lock(mutex);
-	    while (tasks.empty())
-		condition.wait(lock);
-	    Task task = tasks.front();
-	    tasks.pop();
+	    boost::unique_lock<boost::mutex> lock(method_call_mutex);
+	    while (method_call_tasks.empty())
+		method_call_condition.wait(lock);
+	    MethodCallTask method_call_task = method_call_tasks.front();
+	    method_call_tasks.pop();
 	    lock.unlock();
 
-	    dispatch(task.conn, task.msg);
+	    dispatch(method_call_task.conn, method_call_task.msg);
 	}
     }
     catch (const boost::thread_interrupted&)
@@ -1957,11 +2032,11 @@ Clients::find(const string& name)
 
 
 Clients::iterator
-Clients::add(const string& name)
+Clients::add(const string& name, uid_t uid)
 {
     assert(find(name) == entries.end());
 
-    entries.emplace_back(name, *this);
+    entries.emplace_back(name, uid, *this);
 
     return --entries.end();
 }
@@ -1972,7 +2047,7 @@ Clients::remove_zombies()
 {
     for (iterator it = begin(); it != end();)
     {
-	if (it->zombie && it->thread.timed_join(boost::posix_time::seconds(0)))
+	if (it->zombie && it->method_call_thread.timed_join(boost::posix_time::seconds(0)))
 	    it = entries.erase(it);
 	else
 	    ++it;
@@ -1989,3 +2064,36 @@ Clients::has_zombies() const
 
     return false;
 }
+
+
+void
+Client::files_transfer_worker()
+{
+    try
+    {
+	while (true)
+	{
+	    boost::unique_lock<boost::mutex> lock(files_transfer_mutex);
+	    while (files_transfer_tasks.empty())
+		files_transfer_condition.wait(lock);
+
+	    shared_ptr<FilesTransferTask> ptr(files_transfer_tasks.front());
+	    files_transfer_tasks.pop();
+	    lock.unlock();
+
+	    try
+	    {
+		ptr->run();
+	    }
+	    catch (const StreamException& e)
+	    {
+		SN_CAUGHT(e);
+		y2err("error occured during files transfer");
+	    }
+	}
+    }
+    catch (const boost::thread_interrupted&)
+    {
+	y2deb("files transfer worker interrupted");
+    }
+}
